import asyncio
import os
import json
from concurrent.futures import ProcessPoolExecutor
from trellio.host import Host
from urllib.request import urlopen
import hashlib

class ConfigClient:#assign it, at the end of class

    def __init__(self, host_ip, host_port, file_path, h_service, t_service):
        self.host_ip = host_ip
        self.host_port = host_port
        self.config_file = file_path
        self._loop = asyncio.get_event_loop()
        self.http_service = h_service
        self.tcp_service = t_service
        self.config_host_reader = None
        self.config_host_writer = None
        self.running = False
        self._run_lock1 = asyncio.Lock()
        self._run_lock2 = asyncio.Lock()
        self.data = {}

    def start_config(self):
        self.export_to_host()

    def export_to_host(self):
        yield self._connect_to_config_host()#creates connection to config host
        d = {'FILE_PATH': self.config_file}
        self._upload(json.dumps(d))#upload the file path to config host


    def get_file_content(self):
        try:
            #remote file
            data = urlopen(self.config_file).read()#blocking, we have to wait for file
            return json.loads(data)
        except:
            #local file
            if os.path.exists(self.config_file):
                with open(self.config_file, 'rb') as file:
                    data = file.read()
                    self.data = json.loads(data)
                    return data
            else:
                raise Exception('Invalid file details!!')


    async def _connect_to_config_host(self):
        if not(self.config_host_reader or self.config_host_writer):
            client_reader, client_writer = await asyncio.open_connection(self.host_ip,
                                                                      self.host_port)
            self.config_host_reader = client_reader
            self.config_host_writer = client_writer

    def _upload(self, json_string):
        self.config_host_writer.write(json_string)
        yield from self.config_host_writer.drain()


    def handle_request(self, req_str):
        if req_str == ':start_service':
            self.restart()

    def config_handler(self):#poll on file and config host will directly change the file
        def _poll(reader):
            while True:#optimize
                data = yield from reader.read()
                if data:
                    self.handle_request(data)
                else:
                    yield from asyncio.sleep(1)

        file_hash = hashlib.md5(self.get_file_content()).hexdigest()

        def _poll_on_file():
            while True:
                if file_hash != hashlib.md5(self.get_file_content()).hexdigest():
                    self.restart()
                else:
                    yield from asyncio.sleep()

        infinite_future1 = self._loop.run_in_executor(ProcessPoolExecutor(max_workers=1),
                                                     _poll,
                                                     self.config_host_reader)
        infinite_future2 = self._loop.run_in_executor(ProcessPoolExecutor(max_workers=1),
                                                     _poll_on_file)

    def restart(self):
        with (yield from self._run_lock1):
            self._stop()
            self.running = False
        with (yield from self._run_lock2):
            self._run()
            self.running = True

    def _run(self):#incomplete
        '''
        {
          "HTTP_SERVICE_HOST": "",
          "HTTP_SERVICE_PORT": "",
          "TCP_SERVICE_HOST": "",
          "TCP_SERVICE_PORT": "",
          "REGISTRY_HOST": "",
          "REGISTRY_PORT": "",
          "UNIQUE_SERVICE_NAME": "",
          "PUBSUB_H0ST": "",
          "PUBSUB_PORT": ""
        }
        :return:
        '''
        if not self.data:
            self.data = self.get_file_content()

        Host.name = self.data['UNIQUE_SERVICE_NAME']
        if self.tcp_service:
            tcp = self.tcp_service(self.data['TCP_SERVICE_HOST'], self.data['TCP_SERVICE_PORT'])
            Host.attach_tcp_service(tcp)
        if self.http_service:
            http = self.http_service(self.data['HTTP_SERVICE_HOST'], self.data['HTTP_SERVICE_PORT'])
            Host.attach_http_service(http)
        Host.registry_host = self.data['REGISTRY_HOST']
        Host.registry_port = self.data['REGISTRY_PORT']
        Host.pubsub_host = self.data['PUBSUB_HOST']
        Host.pubsub_port = self.data['PUBSUB_PORT']
        Host.run()

    def _stop(self):
        asyncio.get_event_loop().stop()



import inspect

class BaseSignal:
	_registry_list = []

	@classmethod
	def register(cls, to_register, soft=False):
		if inspect.iscoroutine(to_register):
			cls._registry_list.append([to_register, soft])


	@classmethod
	async def _run(cls, host_class):
		raise Warning("Do not call it directly!!")
		for i in cls._registry_list:
			try:
				await i[0](host_class)
			except Exception as e:
				if not i[1]:
					raise e

class ServiceReady(BaseSignal):
	pass


# responsible for storage of all config and connections to config clients
import os
import sys
import json
import asyncio
import asyncio.streams
from aiohttp import web
from urllib.request import urlopen


class ConfigHostAdmin:
    def __init__(self, config_host):
        self.admin_http_host = ''  # host addr at which confighost web api will be served
        self.admin_http_port = ''  # port
        self.config_host = config_host

    def get_aiohttp_application(self):
        def handle_client_info(request):
            name = request.match_info.get('name')
            name_dict = {i['service_name']: i for i in self.config_host.CLIENT_CONNECTIONS}
            if name in name_dict:
                return web.Response(text=str(name_dict[name]))  # placeholder

        app = web.Application()
        app.router.add_get('/service/{name}/', handle_client_info)
        return app

    def start(self):
        self.app = self.get_aiohttp_application()
        web.run_app(self.app, host=self.admin_http_host, port=self.admin_http_port)

    def stop(self):
        pass


def config_admin_factory(c_h):
    return ConfigHostAdmin(c_h)


class BaseConfigServer:  # over-ride usecase, creating a centralized config server on a domain using nginx etc.
    '''
    this class does'nt have logic for config management
    '''

    def __init__(self):
        self.__clients = {}

    def _accept_client(self, client_reader, client_writer):
        """
        This method accepts a new config connection
        """
        # start a new Task to handle this specific client connection
        task = asyncio.Task(self._handle_client(client_reader, client_writer))
        self.__clients[task] = (client_reader, client_writer)  # temporary dictionary until validation is done

        def client_done(task):
            self.client_handler_exit()
            del self.__clients[task]

        task.add_done_callback(client_done)

    async def _handle_client(self, client_reader, client_writer):
        data = await client_reader.read().decode("utf-8")
        await self.config_handler(data, client_reader, client_writer)
        # This enables us to have flow control in our connection.
        await client_writer.drain()

    def config_handler(self, data, client_reader, client_writer):  # method where everything is done
        raise NotImplementedError

    def client_exit(self):
        raise NotImplementedError

    def client_handler_exit(self):
        raise NotImplementedError

    def start(self, loop, host, port):
        self.server = loop.run_until_complete(
            asyncio.streams.start_server(self._accept_client,
                                         host, port,
                                         loop=loop))

    def stop(self, loop):
        if self.server is not None:
            self.server.close()
            loop.run_until_complete(self.server.wait_closed())
            self.server = None


class ConfigHost(BaseConfigServer):
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self.loop = asyncio.get_event_loop()
        self.CLIENT_CONNECTIONS = {}
        self.admin = config_admin_factory(self)
        super(ConfigHost, self).___init__(host, port)

    async def config_handler(self, data, client_reader, client_writer):  # main method
        file_path = json.loads(data)
        data = self.get_file_content(file_path['FILE_PATH'])
        client_dict = {}
        client_dict['raw_config'] = data
        client_dict['reader'] = client_reader
        client_dict['writer'] = client_writer
        self.CLIENT_CONNECTIONS[data['UNIQUE_SERVICE_NAME']] = client_dict  # service name will be unique
        await self.start_client(data['UNIQUE_SERVICE_NAME'])  # start the service now

    async def start_client(self, service_name):
        c_d = self.CLIENT_CONNECTIONS[service_name]
        c_d['writer'].write(':start_client')
        await c_d['writer'].drain()

    def get_file_content(self, c_file):
        try:
            # remote file
            data = urlopen(c_file).read()  # blocking, we have to wait for file
            return json.loads(data)
        except:
            # local file
            if os.path.exists(c_file):
                with open(c_file, 'rb') as file:
                    data = file.read()
                    self.data = json.loads(data)
                    return data
            else:
                raise Exception('Invalid file details!!')

    async def update_client_file(self, service_name):  # will be called by web api, cli command etc.,
        # when file is changed successfully
        new_data = self.get_file_content(self.CLIENT_CONNECTIONS[service_name]['FILE_PATH'])
        n_service_name = new_data['UNIQUE_SERVICE_NAME']
        self.CLIENT_CONNECTIONS[n_service_name]['raw_config'] = new_data
        if service_name != n_service_name:
            del self.CLIENT_CONNECTIONS[service_name]

    def start_config(self):
        self.start(self.loop, self._host, self._port)
        if self.admin:
            self.admin.start()

    def stop_config(self):
        if self.admin:
            self.admin.stop()
        self.stop(asyncio.get_event_loop())
